/*
* kino_common.cc KINO GUI Common Object
* Copyright (C) 2001 Charles Yates <charles.yates@pandora.be>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <fstream>
#include <iostream>
using std::cout;
using std::endl;

#include <string>

#include <sys/types.h>
#include <dirent.h>
#include <libgen.h>
#include <string.h>
#include <stdarg.h>

#include "kino_common.h"
#include "page_editor.h"
#include "page_capture.h"
#include "page_timeline.h"
#include "page_undefined.h"
#include "page_export.h"
#include "page_bttv.h"
#include "page_trim.h"
#include "page_magick.h"
#include "preferences.h"
#include "message.h"
#include "frame.h"
#include "displayer.h"
#include "storyboard.h"
#include "preferences.h"
#include "riff.h"
#include "avi.h"
#include "filehandler.h"
#include "ieee1394io.h"
#include "error.h"


extern "C"
{
#include <pthread.h>
}

/** Constructor for KinoCommon - initialises all GUI widgets.
 
  	\param widget top level widget for the main window
*/

KinoCommon::KinoCommon( GtkWidget *widget ) : last_directory( "" )
{
	cout << "> Kino Common being built" << endl;

	// Initialise class variables from GUI components
	this->widget = widget;
	this->edit_menu = lookup_widget( widget, "edit" );
	this->command_menu = lookup_widget( widget, "commands" );
	this->view_menu = lookup_widget( widget, "view1" );
	this->notebook = GTK_NOTEBOOK( lookup_widget( widget, "main_notebook" ) );
	this->command = GTK_ENTRY( lookup_widget( widget, "command" ) );
	this->video_start_movie_button = GTK_BUTTON( lookup_widget( widget, "video_start_movie_button" ) );
	this->video_start_scene_button = GTK_BUTTON( lookup_widget( widget, "video_start_scene_button" ) );
	this->video_rewind_button = GTK_BUTTON( lookup_widget( widget, "video_rewind_button" ) );
	this->video_back_button = GTK_BUTTON( lookup_widget( widget, "video_back_button" ) );
	this->video_play_button = GTK_BUTTON( lookup_widget( widget, "video_play_button" ) );
	this->video_pause_button = GTK_BUTTON( lookup_widget( widget, "video_pause_button" ) );
	this->video_stop_button = GTK_BUTTON( lookup_widget( widget, "video_stop_button" ) );
	this->video_forward_button = GTK_BUTTON( lookup_widget( widget, "video_forward_button" ) );
	this->video_fast_forward_button = GTK_BUTTON( lookup_widget( widget, "video_fast_forward_button" ) );
	this->video_end_scene_button = GTK_BUTTON( lookup_widget( widget, "video_end_scene_button" ) );
	this->video_end_movie_button = GTK_BUTTON( lookup_widget( widget, "video_end_movie_button" ) );
	this->video_shuttle = GTK_RANGE( lookup_widget( widget, "hscale_shuttle" ) );
	this->statusbar = GTK_STATUSBAR( lookup_widget( widget, "statusbar" ) );

	// Initialise other class variables
	this->currentPage = -1;
	//strcpy( this->currentDirectory, Preferences::getInstance().defaultDirectory );
	strcpy( this->tempFileName, "" );
	//strcpy( this->playlistFileName, "Untitled" );
	// strcpy( this->playlistName, "" );

	// Create page objects
	this->editor = new PageEditor( this );
	this->capture = new PageCapture( this );
	this->timeline = new PageTimeline( this );
	this->undefined = new PageUndefined( this );
	this->bttv = new PageBttv( this );
	this->exportPage = new PageExport( this );
	this->trimPage = new PageTrim( this );
	this->magickPage = new PageMagick( this );

	this->g_currentFrame = -1;
	this->hasListChanged = TRUE;
	this->currentScene = -1;

	// Set to the first page
	//this->setCurrentPage( PAGE_EDITOR );
	this->currentPage = -1;

	this->component_state = VIDEO_STOP;
	this->is_component_state_changing = false;
	if ( Preferences::getInstance().enableTimecode )
		this->timeFormat = TIME_FORMAT_SMPTE;
	else
		this->timeFormat = TIME_FORMAT_NONE;
}

/**	Destructor for the kino common object.
 */

KinoCommon::~KinoCommon()
{
	cout << "> Kino Common being destroyed" << endl;
	clean();
	delete this->editor;
	delete this->capture;
	delete this->timeline;
	delete this->undefined;
	delete this->bttv;
	delete this->exportPage;
	delete this->trimPage;
	delete this->magickPage;
	GetFileMap() ->Clear();
	cout << "> Kino Common destroyed" << endl;
}

/** Carries out a New File request in the kino application. All notebook
	pages are informed of the action through their own newFile method.
*/

bool KinoCommon::newFile( )
{
	bool result = true;

	cout << ">> Kino Common newFile" << endl;

	this->changePageRequest( 0 );
	if ( getPlayList() ->IsDirty() )
	{
		GtkWidget * dialog = gtk_message_dialog_new(
		                         GTK_WINDOW( this->getWidget() ),
		                         GTK_DIALOG_DESTROY_WITH_PARENT,
		                         GTK_MESSAGE_QUESTION,
		                         GTK_BUTTONS_YES_NO,
		                         _( "Project is not saved - do you wish to do so now?" ) );

		switch ( gtk_dialog_run( GTK_DIALOG( dialog ) ) )
		{
		case GTK_RESPONSE_YES:
			result = savePlayList( );
			break;
		case GTK_RESPONSE_NO:
			result = true;
			break;
		default:
			// Do nothing - action cancelled
			result = false;
			break;
		}
		gtk_widget_destroy( dialog );
	}
	if ( result )
	{
		// Specifics for this class
		vector <string> list;
		PlayList clean;

		getPlayList( ) ->GetLastCleanPlayList( clean );
		GetFileMap() ->GetUnusedFxFiles( clean, list );

		if ( list.begin() != list.end() )
		{
			string buffer;

			buffer = _( "The following fx rendered files are no longer used: \n\n" );
			vector < string >::iterator it;
			for ( it = list.begin(); it != list.end(); it ++ )
				buffer += *it + "\n";
			buffer += _( "\nDo you want me to remove them now?" );

			GtkWidget *dialog = gtk_message_dialog_new (
			                        GTK_WINDOW( this->getWidget() ),
			                        GTK_DIALOG_DESTROY_WITH_PARENT,
			                        GTK_MESSAGE_QUESTION,
			                        GTK_BUTTONS_YES_NO,
			                        buffer.c_str() );

			switch ( gtk_dialog_run( GTK_DIALOG( dialog ) ) )
			{
			case GTK_RESPONSE_YES:
				for ( it = list.begin(); it != list.end(); it ++ )
					unlink( ( *it ).c_str() );
				result = true;
				break;
			case GTK_RESPONSE_NO:
				// Do nothing - leave files on the system
				result = true;
				break;
			default:
				// Do nothing - action cancelled
				result = false;
				break;
			}
			gtk_widget_destroy( dialog );
		}

		if ( result )
		{
			getPlayList() ->CleanPlayList( );
			GetFileMap() ->Clear();
			this->g_currentFrame = -1;
			this->currentScene = -1;
			this->setWindowTitle( );

			// Call new File for each page
			Page *page = NULL;
			for ( int index = 0; ( page = getPage( index ) ) != NULL; index ++ )
				page->newFile();

			GetStoryboard() ->reset();
			GetStoryboard() ->redraw();
			this->windowMoved();

			GetEditorBackup() ->Clear();
			getPageEditor() ->snapshot();
		}
	}

	return result;
}

/**	Sends change page request to the gui.
 
	\param page page to change to
*/

void KinoCommon::changePageRequest( int page )
{
	gtk_notebook_set_page( notebook, page );
}

/**	Carries out a change page - this should be triggered by changePageRequest
	to ensure that the GUI stays in sync with the common structure.
 
	\param page page to change to
*/

void KinoCommon::setCurrentPage( int page )
{
	if ( this->currentPage != page )
	{
		if ( this->currentPage != -1 )
			clean();
		this->currentPage = page;
		start();
		moveToFrame();
		gtk_widget_queue_draw( getWidget() );
	}
}

/** Returns the Page object associated to the specified notebook page.
 
	\param page index of Page object to return
	\return the Page object associated to the given notebook page or NULL if invalid.
*/

Page *KinoCommon::getPage( int page )
{
	Page * ret = NULL;

	switch ( page )
	{
	case PAGE_EDITOR:
		ret = getPageEditor();
		break;
	case PAGE_CAPTURE:
		ret = getPageCapture();
		break;
	case PAGE_TIMELINE:
		ret = getPageTimeline();
		break;
	case PAGE_TRIM:
		ret = getPageTrim();
		break;
	case PAGE_EXPORT:
		ret = getPageExport();
		break;
	case PAGE_MAGICK:
		ret = getPageMagick();
		break;
	case PAGE_BTTV:
		ret = getPageBttv();
		break;
	}

	return ret;
}

/** Returns the current Page object.
 
	\return the current Page object (will be PageUndefined if current page is invalid).
*/

Page *KinoCommon::getCurrentPage( )
{
	Page * ret = getPage( this->currentPage );
	if ( ret == NULL )
		ret = getPageUndefined();
	return ret;
}

/** Returns the Page associated to the Editor page.
 
	\return the Page object requested
*/

PageEditor *KinoCommon::getPageEditor( )
{
	return editor;
}

/** Returns the Page associated to the Capture page.
 
	\return the Page object requested
*/

PageCapture *KinoCommon::getPageCapture( )
{
	return capture;
}

/** Returns the Page associated to the Timeline page.
 
	\return the Page object requested
*/

PageTimeline *KinoCommon::getPageTimeline( )
{
	return timeline;
}

/** Returns the Page associated to the Undefined page.
 
	\return the Page object requested
*/

PageUndefined *KinoCommon::getPageUndefined( )
{
	return undefined;
}

/** Returns the 'BTTV' page (test code for extending hardware support).
 
	\return the Page object requested
*/

PageBttv *KinoCommon::getPageBttv( )
{
	return bttv;
}

/** Returns the Export page.
 
	\return the Page object requested
*/

PageExport *KinoCommon::getPageExport( )
{
	return exportPage;
}

/** Returns the Trim page.
 
	\return the Page object requested
*/

PageTrim *KinoCommon::getPageTrim( )
{
	return trimPage;
}

/** Returns the Magick page.
 
	\return the Page object requested
*/

PageMagick *KinoCommon::getPageMagick( )
{
	return magickPage;
}

/** Inserts a file at the current position.
*/

void KinoCommon::insertFile( )
{
	char * filename = this->getFileToOpen( _( "Choose a DV or SMIL file to insert" ) );
	if ( strcmp( filename, "" ) )
	{
		switch ( checkFile( filename ) )
		{
		case AVI:
		case RAW_DV:
		case QT:
			if ( loadMediaObject( filename, g_currentFrame == -1 ? 0 : g_currentFrame ) )
			{
				hasListChanged = TRUE;
				getPlayList() ->SetDirty( true );
			}
			break;
		case PLAYLIST:
			if ( loadPlayList( filename, g_currentFrame ) )
			{
				hasListChanged = TRUE;
				getPlayList() ->SetDirty( true );
				getPageEditor() ->snapshot();
			}
			break;
		case UNKNOWN_FORMAT:
			modal_message( _( "Invalid file specified." ) );
			break;
		}
	}

	setWindowTitle();
	moveToFrame( );
}

/** Appends a file after the current position.
*/

void KinoCommon::appendFile( )
{
	char * filename = this->getFileToOpen( _( "Choose a DV or SMIL file to append" ) );
	if ( strcmp( filename, "" ) )
	{
		switch ( checkFile( filename ) )
		{
		case AVI:
		case RAW_DV:
		case QT:
			if ( loadMediaObject( filename, g_currentFrame + 1 ) )
			{
				g_currentFrame++;
				hasListChanged = TRUE;
				getPlayList() ->SetDirty( true );
			}
			break;
		case PLAYLIST:
			// Insert the new smil
			if ( loadPlayList( filename, g_currentFrame + 1 ) )
			{
				g_currentFrame++;
				hasListChanged = TRUE;
				getPlayList() ->SetDirty( true );
				getPageEditor() ->snapshot();
			}
			break;
		case UNKNOWN_FORMAT:
			modal_message( _( "Invalid file specified." ) );
			break;
		}
	}

	setWindowTitle();
	moveToFrame( );
}

/** Opens a file. An AVI or a PLAYLIST causes a newFile to called prior to the load
	and a PLAYLIST becomes the currently editted file.
*/

void KinoCommon::loadFile( )
{
	char * filename = this->getFileToOpen( _( "Choose a DV or SMIL file" ) );

	if ( strcmp( filename, "" ) )
	{
		switch ( checkFile( filename ) )
		{
		case AVI:
		case RAW_DV:
		case QT:
			if ( newFile( ) )
			{
				GetEditorBackup() ->Clear();
				if ( loadMediaObject( filename, 0 ) )
				{
					g_currentFrame = 0;
					hasListChanged = TRUE;
				}
			}
			break;
		case PLAYLIST:
			if ( newFile( ) )
			{
				GetEditorBackup() ->Clear();
				if ( loadPlayList( filename, 0 ) )
				{
					g_currentFrame = 0;
					hasListChanged = TRUE;
					getPageEditor( ) ->snapshot( );
					getPlayList( ) ->SetDocName( filename );
					getPlayList( ) ->SetDirty( false );
				}
			}
			break;
		case UNKNOWN_FORMAT:
			modal_message( _( "Invalid file specified." ) );
			break;
		}
	}
	setWindowTitle();
	this->currentScene = 0;
	moveToFrame( );
}

/** Save the current play list with a different name to the original loaded (or
	previously saved as) play list file name. 
*/

bool KinoCommon::savePlayListAs( )
{
	bool result = false;
	changePageRequest( 0 );
	char *filename = this->getFileToSave( _( "Save the movie as SMIL" ) );
	if ( strcmp( filename, "" ) )
		result = savePlayList( filename );
	return result;
}

/** Save the current play list with a different name to the original loaded (or
       previously saved as) play list file name in Eli format
*/

void KinoCommon::savePlayListAsEli( )
{
	Frame frame;

	if ( g_currentFrame == -1 )
		return ;

	this->getPlayList() ->GetFrame( g_currentFrame, frame );
	changePageRequest( 0 );
	char *filename = this->getFileToSave( _( "Save the movie as ELI" ) );
	if ( strcmp( filename, "" ) )
	{
		if ( this->getPlayList() ->SavePlayListEli( filename,
		        frame.IsPAL() ) == false )
			modal_message( _( "Could not save the ELI" ) );
	}
}

/** Save the current play list to the file used to load the playlist originally. If
	no file was originally used then use savePlayListAs to obtain one.
*/

bool KinoCommon::savePlayList( )
{
	bool result = true;
	if ( getPlayList() ->GetDocName() == "" )
		result = savePlayListAs( );
	else
		result = savePlayList( ( char * ) ( getPlayList() ->GetDocName().c_str() ) );
	return result;
}


/**	Save the current frame as a still frame.
*/

void KinoCommon::saveFrame( )
{
	getCurrentPage() ->saveFrame();
}

/** Handles the modal dialog for collecting a file to open. The file is not opened,
	but returned by this method.
 
	It relies on the the callback functions defined for the open_movie_dialog. All three
	of these must call gtk_main_quit when finished and the delete and cancel should set
	the filename to an empty string.
 
	\todo remove glade generated code for modal dialogs
	\param title title of the window
	\return a string inidicating the selected file (an empty string denotes no selection)
*/

char *KinoCommon::getFileToOpen( char *title )
{
	changePageRequest( 0 );
	GtkWidget *dialog = gtk_file_selection_new( title );
	gtk_window_set_modal( GTK_WINDOW( dialog ), TRUE );
	gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION( dialog )->ok_button ),
				"clicked", GTK_SIGNAL_FUNC( gtk_widget_hide ),
				(gpointer) dialog );
	gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION( dialog )->cancel_button ),
				"clicked", GTK_SIGNAL_FUNC( gtk_widget_hide ),
				(gpointer) dialog );
	if ( last_directory == "" )
	{
		string directory = getPlayList() ->GetProjectDirectory( );
		if ( directory != "" )
			gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), ( char * ) ( directory + "/" ).c_str() );
		last_directory = directory;
	}
	else
	{
		gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), ( char * ) ( last_directory + "/" ).c_str() );
	}
	gtk_dialog_run( GTK_DIALOG( dialog ) );
	const gchar *filename = gtk_file_selection_get_filename( GTK_FILE_SELECTION( dialog ) );
	/* If the user cancels, a directory will still be returned
	   If this is the case, return "" - the caller knows how to handle this */
	if ( opendir( filename ) == NULL )
	{
		strcpy( tempFileName, filename );
		last_directory = directory_utils::get_directory_from_file( filename );
	}
	else
		strcpy( tempFileName, "" );
	gtk_widget_destroy( dialog );
	return tempFileName;
}

/** Handles the modal dialog for collecting a file to save. The file is not saved,
	but returned by this method.
 
	It relies on the the callback functions defined for the save_movie_dialog. All three
	of these must call gtk_main_quit when finished and the delete and cancel should set
	the filename to an empty string.
 
	\todo remove glade generated code for modal dialogs
	\param title title of the window
	\return a string inidicating the selected file (an empty string denotes no selection)
*/

char *KinoCommon::getFileToSave( char *title )
{
	changePageRequest( 0 );
	GtkWidget *dialog = gtk_file_selection_new( title );
	gtk_window_set_modal( GTK_WINDOW( dialog ), TRUE );
	gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION( dialog )->ok_button ),
				"clicked", GTK_SIGNAL_FUNC( gtk_widget_hide ),
				(gpointer) dialog );
	gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION( dialog )->cancel_button ),
				"clicked", GTK_SIGNAL_FUNC( gtk_widget_hide ),
				(gpointer) dialog );
	if ( last_directory == "" )
	{
		string directory = getPlayList() ->GetProjectDirectory( );
		if ( directory != "" )
			gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), ( char * ) ( directory + "/" ).c_str() );
		last_directory = directory;
	}
	else
	{
		gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), ( char * ) ( last_directory + "/" ).c_str() );
	}
	gtk_dialog_run( GTK_DIALOG( dialog ) );
	const gchar *filename = gtk_file_selection_get_filename( GTK_FILE_SELECTION( dialog ) );
	if ( opendir( filename ) == NULL )
	{
		strcpy( tempFileName, filename );
		last_directory = directory_utils::get_directory_from_file( filename );
	}
	else
		strcpy( tempFileName, "" );
	gtk_widget_destroy( dialog );
	return tempFileName;
}

/**	Determines the format of the file.
 
	It is required that a file be at least 20 bytes for it to load.
 
	Note that it is currently required that mediafiles have either
	a .dv or .avi suffix since the PlayList uses the suffixes .avi
	and .dv to determine what FileHandler to create. Smil files
	are allowed to load without suffixes. (Correct?)
 
	\param filename file to check
	\return AVI, PLAYLIST, RAW_DV or UNKNOWN_FORMAT
*/

int KinoCommon::checkFile( char *FileName )
{
	// Try reading a few bytes from the file ...
	std::ifstream file( FileName );
	std::vector<char> buffer( 22, '\0' );
	file.read( &buffer[ 0 ], buffer.size() );

	// If it didn't work, we're done ...
	if ( file.bad() )
	{
		cerr << "> Error reading file: " << FileName << endl;
		return UNKNOWN_FORMAT;
	}

	// If the file is shorter than our buffer, we're done ...
	if ( file.eof() )
	{
		cerr << "> File size < " << buffer.size() << " bytes: " << FileName << endl;
		return UNKNOWN_FORMAT;
	}

	// Start looking at file suffixes ...
	const std::string filename( FileName );
	const std::string suffix( filename.begin() + filename.rfind( "." ), filename.end() );

	if ( suffix == ".avi" )
	{
		return AVI;
	}
	else if ( suffix == ".dv" || suffix == ".dif" )
	{
		// This bit of magic brought to you from the dvgrab AVI iso handler ...
		const unsigned char * const p = reinterpret_cast<unsigned char*>( &buffer[ 0 ] );
		const int section_type = p[ 0 ] >> 5;
		const int dif_sequence = p[ 1 ] >> 4;

		if ( 0 == section_type && 0 == dif_sequence )
			return RAW_DV;

		return UNKNOWN_FORMAT;
	}
	else if ( suffix == ".mov" )
	{
		return QT;
	}

	// Check to see if it's SMIL ...
	const std::string smil_magic( "<?xml version=\"1.0\"?>" );
	if ( std::string( buffer.begin(), buffer.begin() + smil_magic.size() ) == smil_magic )
		return PLAYLIST;

	// No can do!
	return UNKNOWN_FORMAT;
}

/** Sets the document name in the window title.
*/
void KinoCommon::setWindowTitle( )
{
	GtkWidget * mainWindow = lookup_widget( this->getWidget(), "main_window" );
	if ( mainWindow )
	{
		char strbuf[ 1024 ];
		strcpy( strbuf, "Kino " VERSION ": " );
		//strcat(strbuf, docname);
		if ( getPlayList() ->GetDocName( ) == "" )
			strcat( strbuf, _( "Untitled" ) );
		else
			strcat( strbuf, getPlayList() ->GetDocName( ).c_str() );
		if ( getPlayList( ) ->IsDirty( ) )
			strcat( strbuf, " * " );
		gtk_window_set_title( GTK_WINDOW( mainWindow ), strbuf );
	}
}

/** Loads the specified media file before the specified frame.
 
	\param file file to load
	\param before frame to insert at
*/

bool KinoCommon::loadMediaObject( char *file, int before )
{

	PlayList newList;
	bool result = true;

	try
	{
		if ( newList.LoadMediaObject( file ) )
		{
			this->getPlayList() ->InsertPlayList( newList, before );
			getPageEditor() ->snapshot();
		}
		else
		{
			modal_message( _( "Failed to load media file." ) );
			result = false;
		}
	}
	catch ( string s )
	{
		cout << "Could not load file " << file << ", because an exception has occurred: " << endl;
		cout << s << endl;
		result = false;
	}

	return result;
}


/** Loads the specified play list before the specified frame.
 
	\param file file to load
	\param before frame to insert at
*/

bool KinoCommon::loadPlayList( char *file, int before )
{

	PlayList newList;
	bool result = newList.LoadPlayList( file );

	//changeDirectory(file);
	if ( result )
		this->getPlayList() ->InsertPlayList( newList, before );
	else
		modal_message( _( "Could not load the SMIL" ) );

	return result;
}

/** Saves the playlist.
 
	\param file file to save
*/

bool KinoCommon::savePlayList( char *file )
{
	bool result = false;
	result = getPlayList() ->SavePlayList( file );

	if ( result )
		setWindowTitle( );
	else
		modal_message( _( "Could not save the SMIL" ) );

	return result;
}

/** Loads all files specified in the command line style arguments.
 
  	\param argc number of arguments
	\param argv arguments
*/

void KinoCommon::bulkLoad( int argc, char* argv[] )
{

	char filename[ PATH_MAX + NAME_MAX ];
	for ( int i = 1; i < argc; ++i )
	{
		/* Get the real name of the file, to make sure that
		we store absolute paths in smil files */
		if ( NULL != realpath( argv[ i ], filename ) )
		{

			switch ( checkFile( filename ) )
			{
			case AVI:
			case RAW_DV:
			case QT:
				if ( loadMediaObject( filename, this->getPlayList() ->GetNumFrames() ) )
				{
					if ( g_currentFrame == -1 )
						g_currentFrame = 0;
					this->hasListChanged = TRUE;
				}
				else
				{
					cerr << "KinoCommon::bulkLoad: Failed to load " << filename << endl;
				}
				break;
			case PLAYLIST:
				{
					int last_count = getPlayList( ) ->GetNumFrames( );
					string last_doc_name = getPlayList( ) ->GetDocName( );
					if ( loadPlayList( filename, last_count ) )
					{
						if ( last_count == 0 && last_doc_name == "" )
						{
							getPlayList() ->SetDocName( filename );
							getPlayList() ->SetDirty( false );
						}
						else
						{
							getPlayList() ->SetDirty( true );
						}
						getPageEditor() ->snapshot();
						if ( g_currentFrame == -1 )
							g_currentFrame = 0;
						this->hasListChanged = TRUE;
					}
				}
				break;
			case UNKNOWN_FORMAT:
				/* DRD> I had to comment out the following to prevent lockups with
				  invalid args. When Xv crashes my desktop, and I log back in, Kino is
				  started automatically by GNOME which passes some weird --smclient arg */ 
				//				modal_message( "Invalid file specified." );
				cerr << "KinoCommon::bulkLoad: Invalid file specified: "
				<< filename << endl;
				break;
			}
		}
		else
		{
			cerr << "KinoCommon::bulkLoad: Unable to resolve " << filename << endl;
		}
	}

	setWindowTitle( );
}

/** Save the specified frame to the specified file.
 
  	\param	abs_frame_num the frame to save
	\param	file the file to save it to
*/

void KinoCommon::saveFrame( int abs_frame_num, char *file )
{}


/** Relay the current command and any message associated to it.
 
  	\param cmd command
	\param msg message;
*/

void KinoCommon::keyboardFeedback( const char *cmd, const char *msg )
{
	char s[ 256 ];

	strcpy( s, cmd );
	strcat( s, "   " );
	strcat( s, msg );

	setStatusBar( s );
}

/** Move to the current frame. Corrects the frame specified to be within the
	range of the current play list and informs the currently selected page of the
	change.
 
	\param frame requested frame
	\return the frame actually moved to
*/

int KinoCommon::moveToFrame( )
{
	if ( currentPage == PAGE_TRIM )
	{
		getPageTrim() ->movedToFrame( getPageTrim() ->getPosition() );
		return getPageTrim() ->getPosition();
	}
	return moveToFrame( g_currentFrame );
}

/** Move to the frame specified. Corrects the frame specified to be within the
	range of the current play list and informs the currently selected page of the
	change.
 
	\param frame requested frame
	\return the frame actually moved to
*/

int KinoCommon::moveToFrame( int frame )
{
	if ( currentPage == PAGE_TRIM )
	{
		getPageTrim() ->movedToFrame( frame );
		return frame;
	}

	if ( frame >= 0 && frame < getPlayList() ->GetNumFrames() )
		g_currentFrame = frame;
	else if ( frame >= getPlayList() ->GetNumFrames() )
		g_currentFrame = getPlayList() ->GetNumFrames() - 1;
	else if ( frame < 0 && getPlayList() ->GetNumFrames() > 0 )
		g_currentFrame = 0;
	else
		g_currentFrame = -1;

	getCurrentPage() ->movedToFrame( g_currentFrame );
	//	if ( this->hasListChanged )
	//		GetStoryboard()->redraw();

	return g_currentFrame;
}

/** Move the frame relative to the number specified.
 
	\param frames requested frame offset
	\return the frame actually moved to
*/

int KinoCommon::moveByFrames( int frames )
{
	if ( frames == 0 )
		return g_currentFrame;
	int frame = g_currentFrame + frames;
	return moveToFrame( frame );
}

/** Relay the current frame info to the GUI.
 
	\param i current frame
*/

void KinoCommon::showFrameInfo( int i )
{

	if ( timeFormat != TIME_FORMAT_NONE )
		getCurrentPage() ->showFrameInfo( i );
}

/** Display detailed information about this frame and its file.
 
	\param i current frame
*/

void KinoCommon::showFrameMoreInfo( Frame &frame, FileHandler *media )
{
	if ( showMoreInfo )
	{
		GtkBin * bin = GTK_BIN( lookup_widget( widget, "frame_more_info" ) );
		GtkLabel *label = GTK_LABEL( bin->child );
		char s[ 2048 ];

		if ( media != NULL )
		{
			TimeCode tc;
			AudioInfo ainfo;
			string format;

			frame.GetTimeCode( tc );
			frame.GetAudioInfo( ainfo );

			string ext = media->GetExtension();
			if ( ext == ".dv" )
				format = "Raw DV";
			else if ( ext == ".avi" )
			{
				AVIHandler * avi = dynamic_cast< AVIHandler* >( media );
				format = ( avi->GetFormat() == AVI_DV1_FORMAT ) ? "AVI, FOURCC dvsd, Type 1" : "AVI, FOURCC dvsd, Type 2";
				if ( avi->GetOpenDML() )
					format += ", OpenDML";
			}
			else if ( ext == ".mov" )
				format = "Quicktime, FOURCC dvsd";
			else
				format = "unknown";


			snprintf( s, 2048, _( "File Name:   %s\nDV Recording Date: %s\nDV Timecode: %2.2d:%2.2d:%2.2d:%2.2d\nFile Format: %s\nAudio: %d bit, %d KHz, %d samples\nVideo: %d x %d, %s, %s, %2.2f fps" ),
			          media->GetFilename().c_str(),
			          frame.GetRecordingDate().c_str(),
			          tc.hour, tc.min, tc.sec, tc.frame,
			          format.c_str(),
			          ainfo.quantization, ainfo.frequency / 1000, ainfo.samples,
			          frame.GetWidth(), frame.GetHeight(),
			          frame.IsPAL() ? "PAL" : "NTSC",
			          frame.IsWide() ? "16:9" : "4:3",
			          frame.GetFrameRate()
			        );
		}
		else
			sprintf( s, "\n\n\n" );

		gtk_label_set_text( label, s );
	}
}

/** Trigger the start action of the current page. This method is also responsible
	for determining the state of the main page buttons and widgets.
*/

void KinoCommon::start()
{
	getCurrentPage() ->start();
	activateWidgets();

	toggleComponents( getComponentState(), false );
	component_enum pattern = ( component_enum ) ( getCurrentPage() ->activate() ^ getCurrentPage() ->deactivate() );
	if ( pattern & VIDEO_STOP )
		toggleComponents( VIDEO_STOP, true );
	commitComponentState();
}

/** Trigger the clean action of the current page.
*/

void KinoCommon::clean()
{
	getCurrentPage() ->clean();
}

/** Trigger the keyboard action of the current page.
*/

gboolean KinoCommon::processKeyboard( GdkEventKey *event )
{
	return getCurrentPage() ->processKeyboard( event );
}

/** Trigger the command line action of the current page.
*/

gboolean KinoCommon::processCommandLine( const char *cmd )
{
	return getCurrentPage() ->processCommandLine( cmd );
}

/** Trigger a menu command action of the current page. These can be
	either keyboard or command line style strings.
*/

gboolean KinoCommon::processMenuCommand( char *cmd )
{
	return getCurrentPage() ->processMenuCommand( cmd );
}

/** Trigger the select scene action of the current page.
*/

void KinoCommon::selectScene( int scene )
{
	getCurrentPage() ->selectScene( scene );
}

/** Trigger the start of movie action of the current page.
*/

void KinoCommon::videoStartOfMovie()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoStartOfMovie();

		commitComponentState();
	}
}

/** Trigger the previous scene action of the current page.
*/

void KinoCommon::videoPreviousScene()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoPreviousScene();

		commitComponentState();
	}
}

/** Trigger the start of scene action of the current page.
*/

void KinoCommon::videoStartOfScene()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoStartOfScene();

		commitComponentState();
	}
}

/** Trigger the rewind action of the current page.
*/

void KinoCommon::videoRewind()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoRewind();

		commitComponentState();
	}
}

/** Trigger the back action of the current page.
*/

void KinoCommon::videoBack()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoBack();

		commitComponentState();
	}
}

/** Trigger the play action of the current page.
*/

void KinoCommon::videoPlay()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoPlay();

		commitComponentState();
	}
}

/** Trigger the forward action of the current page.
*/

void KinoCommon::videoForward()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoForward();

		commitComponentState();
	}
}

/** Trigger the fast forward action of the current page.
*/

void KinoCommon::videoFastForward()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoFastForward();

		commitComponentState();
	}
}

/** Trigger the next scene action of the current page.
*/

void KinoCommon::videoNextScene()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoNextScene();

		commitComponentState();
	}
}

/** Trigger the end of scene action of the current page.
*/

void KinoCommon::videoEndOfScene()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoEndOfScene();

		commitComponentState();
	}
}

/** Trigger the end movie action of the current page.
*/

void KinoCommon::videoEndOfMovie()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoEndOfMovie();

		commitComponentState();
	}
}

/** Trigger the pause action of the current page.
*/

void KinoCommon::videoPause()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoPause();

		commitComponentState();
	}
}

/** Trigger the stop action of the current page.
*/

void KinoCommon::videoStop()
{
	if ( ! is_component_state_changing )
	{

		getCurrentPage() ->videoStop();

		commitComponentState();
	}
}

/** Bi-directional variable-speed playback
 
	\param angle A number from -7 (fastest reverse) to 7 (fastest forward)
*/

void KinoCommon::videoShuttle( int angle )
{
	if ( ! is_component_state_changing )
	{
		toggleComponents( getComponentState(), false );

		//gtk_range_set_value( this->video_shuttle, angle );
		getCurrentPage() ->videoShuttle( angle );

		commitComponentState();
	}
}

void KinoCommon::windowMoved()
{
	getCurrentPage() ->windowMoved();
}

void KinoCommon::visibilityChanged( gboolean visible )
{
	getCurrentPage() ->visibilityChanged( visible );
}

/** Activate or deactivate the widgets at the request of the previous page.
    This method is always called immediately after the current pages start method
	and can be called at the discretion of the page if required.
*/

void KinoCommon::activateWidgets()
{
	component_enum pattern = ( component_enum ) ( getCurrentPage() ->activate() ^ getCurrentPage() ->deactivate() );
	gtk_widget_set_sensitive( GTK_WIDGET( edit_menu ), pattern & EDIT_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_undo" ), pattern & EDIT_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_redo" ), pattern & EDIT_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_cut" ), pattern & EDIT_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_copy" ), pattern & EDIT_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_paste" ), pattern & EDIT_MENU );
	gtk_widget_set_sensitive( GTK_WIDGET( command_menu ), pattern & COMMAND_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_insert" ), pattern & COMMAND_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_append" ), pattern & COMMAND_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_split" ), pattern & COMMAND_MENU );
	gtk_widget_set_sensitive( lookup_widget( widget, "button_join" ), pattern & COMMAND_MENU );
	GetStoryboard() ->setSensitive( pattern & SCENE_LIST );
	gtk_widget_set_sensitive( GTK_WIDGET( video_start_movie_button ), pattern & VIDEO_START_OF_MOVIE );
	gtk_widget_set_sensitive( GTK_WIDGET( video_start_scene_button ), pattern & VIDEO_START_OF_SCENE );
	gtk_widget_set_sensitive( GTK_WIDGET( video_rewind_button ), pattern & VIDEO_REWIND );
	gtk_widget_set_sensitive( GTK_WIDGET( video_back_button ), pattern & VIDEO_BACK );
	gtk_widget_set_sensitive( GTK_WIDGET( video_play_button ), pattern & VIDEO_PLAY );
	gtk_widget_set_sensitive( GTK_WIDGET( video_pause_button ), pattern & VIDEO_PAUSE );
	gtk_widget_set_sensitive( GTK_WIDGET( video_stop_button ), pattern & VIDEO_STOP );
	gtk_widget_set_sensitive( GTK_WIDGET( video_forward_button ), pattern & VIDEO_FORWARD );
	gtk_widget_set_sensitive( GTK_WIDGET( video_fast_forward_button ), pattern & VIDEO_FAST_FORWARD );
	gtk_widget_set_sensitive( GTK_WIDGET( video_end_scene_button ), pattern & VIDEO_NEXT_SCENE );
	gtk_widget_set_sensitive( GTK_WIDGET( video_end_movie_button ), pattern & VIDEO_END_OF_MOVIE );
	gtk_widget_set_sensitive( GTK_WIDGET( video_shuttle ), pattern & VIDEO_SHUTTLE );
	gtk_widget_set_sensitive( GTK_WIDGET( command ), pattern & VIDEO_COMMAND );
}


/** Set the state of toggle buttons.
 
	The state of the buttons are retained in memory, and
	this method will set a flag that the state has changed (dirty),
	but has not yet been committed. The button commands will
	not respond while the state is dirty. Use commitComponentState()
	to commit the changes and let the buttons issue their commands once again.
 
    \param pattern A set of component_enums to toggle
    \param state If true then lower the button, else raise the button
*/

void KinoCommon::toggleComponents( component_enum pattern, bool state )
{

	this->is_component_state_changing = true;

	if ( state )
		this->component_state |= pattern;
	else
		this->component_state ^= pattern;

	if ( pattern & VIDEO_START_OF_MOVIE )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_start_movie_button ), state );
	if ( pattern & VIDEO_START_OF_SCENE )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_start_scene_button ), state );
	if ( pattern & VIDEO_REWIND )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_rewind_button ), state );
	if ( pattern & VIDEO_BACK )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_back_button ), state );
	if ( pattern & VIDEO_PLAY )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_play_button ), state );
	if ( pattern & VIDEO_PAUSE )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_pause_button ), state );
	if ( pattern & VIDEO_STOP )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_stop_button ), state );
	if ( pattern & VIDEO_FORWARD )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_forward_button ), state );
	if ( pattern & VIDEO_FAST_FORWARD )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_fast_forward_button ), state );
	if ( pattern & VIDEO_NEXT_SCENE )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_end_scene_button ), state );
	if ( pattern & VIDEO_END_OF_MOVIE )
		gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( video_end_movie_button ), state );

}

/** Get the current component state data
*/
component_enum KinoCommon::getComponentState()
{
	return ( component_enum ) this->component_state;
}

/** Set the component state data and commit it.
 
    You can set component state at the same time too.
    See the toggleComponents() method.
	
    \param pattern A set of component_enums to set true
*/
void KinoCommon::commitComponentState( component_enum pattern )
{
	this->component_state |= pattern;
	this->is_component_state_changing = false;
}


/** Resize the video preview area to a percentage of source image size.
 
    \param factor The scaling factor from 0 to 1. If zero, then auto-size.
                  If negative, then cycle to the next preset factor.
    \param noWarning A boolean to indicate whether to suppress the warning
                     dialog when Kino is unable to determine video size.
*/

static Frame frame;

void KinoCommon::setPreviewSize( float factor, bool noWarning )
{
	// handle the cycle case
	if ( factor < 0 )
	{
		factor = Preferences::getInstance().previewSize;
		factor += 0.5;
		if ( factor > 1.0 )
			factor = 0.0;
	}
	// save the new size
	Preferences::getInstance().previewSize = factor;

	// handle fixed size factors
	if ( Preferences::getInstance().displayFixed != ( factor > 0 ) )
	{
		Preferences::getInstance().displayFixed = ( factor > 0 );
		packIt( "packer_edit", "packer_edit_outer" );
		packIt( "packer_capture", "packer_capture_outer" );
		packIt( "packer_trim", "packer_trim_outer" );
	}

	if ( factor == 0 )
		factor = 0.5;
		
	if ( this->g_currentFrame != -1 )
	{
		this->getPlayList() ->GetFrame( this->g_currentFrame, frame );
		int width = frame.GetWidth();
		if ( frame.IsWide() )
			width = frame.IsPAL() ? 1024 : 854;
		AspectRatioCalculator calc( width, frame.GetHeight(),
		                            frame.GetWidth(), frame.GetHeight(),
		                            frame.IsPAL(), frame.IsWide() );

		calc.width = ( int ) ( calc.width * factor );
		calc.height = ( int ) ( calc.height * factor );
		cerr << "> setting video preview size to " << calc.width << "x" << calc.height << endl;

		GtkWidget *frameArea = lookup_widget( getWidget(), "packer_edit" );
		gtk_widget_set_size_request( frameArea, calc.width + 4, calc.height + 4 );
		frameArea = lookup_widget( getWidget(), "packer_capture" );
		gtk_widget_set_size_request( frameArea, calc.width + 4, calc.height + 4 );
		frameArea = lookup_widget( getWidget(), "packer_trim" );
		gtk_widget_set_size_request( frameArea, calc.width + 4, calc.height + 4 );

		//gtk_window_resize( GTK_WINDOW( getWidget() ), 1, 1 );
	}
	else if ( Preferences::getInstance().defaultNormalisation != NORM_UNSPECIFIED )
	{
		int width = 720;
		int height = Preferences::getInstance().defaultNormalisation == NORM_PAL ? 576 : 480;
		bool isPAL = Preferences::getInstance().defaultNormalisation == NORM_PAL;
		bool isWide = Preferences::getInstance().defaultAspect == ASPECT_169;
		AspectRatioCalculator calc( width, height, width, height,
		                            isPAL, isWide );
		                            
		calc.width = ( int ) ( calc.width * factor );
		calc.height = ( int ) ( calc.height * factor );
		cerr << "> setting video preview size to " << calc.width << "x" << calc.height << endl;
		                            
		GtkWidget *frameArea = lookup_widget( getWidget(), "packer_edit" );
		gtk_widget_set_size_request( frameArea, calc.width + 4, calc.height + 4 );
		frameArea = lookup_widget( getWidget(), "packer_capture" );
		gtk_widget_set_size_request( frameArea, calc.width + 4, calc.height + 4 );
		frameArea = lookup_widget( getWidget(), "packer_trim" );
		gtk_widget_set_size_request( frameArea, calc.width + 4, calc.height + 4 );

		//gtk_window_resize( GTK_WINDOW( getWidget() ), 1, 1 );
	}
	else if ( ! noWarning )
	{
		modal_message( _( "The playlist is empty and the default preferences for video creation have not been specified." ) );
	}
}

void KinoCommon::loadSplash( GtkWidget *drawable )
{
	if ( GDK_IS_DRAWABLE( widget->window ) )
	{
		gdk_draw_rectangle ( widget->window, widget->style->black_gc, TRUE,
		                     widget->allocation.x, widget->allocation.y,
		                     widget->allocation.width, widget->allocation.height );
	}
}

/** Load the splash image into the video preview area.
 
    \param widget the GtkDrawingArea into which to render the image.
*/
void KinoCommon::loadSplash( GtkDrawingArea *drawable )
{
#if 1
	clearPreview( drawable );
#elif 0

	unsigned char *background;
	int width = ( ( GtkWidget* ) drawable ) ->allocation.width;
	int height = ( ( GtkWidget* ) drawable ) ->allocation.height;
	static bool warnOnce = true;
	GdkGC *gc = gdk_gc_new( ( ( GtkWidget* ) drawable ) ->window );

	background = ( unsigned char * ) malloc( width * height * 3 );
	fail_null( background );
	memset( background, 0, width * height * 3 );
	GdkPixbuf *bgImage = gdk_pixbuf_new_from_data( background, GDK_COLORSPACE_RGB, FALSE, 8,
						 width, height, width * 3, NULL, NULL );
	fail_null( bgImage );
	GdkPixbuf *splash = create_pixbuf( "kino/kino.jpeg" );
	if ( splash != NULL )
	{
		AspectRatioCalculator calc( width, height, gdk_pixbuf_get_width( splash ),
									gdk_pixbuf_get_height( splash ), gdk_pixbuf_get_width( splash ),
									gdk_pixbuf_get_height( splash ) );
		GdkPixbuf *image = gdk_pixbuf_scale_simple( splash, calc.width, calc.height, GDK_INTERP_BILINEAR );
		fail_null( image );

		gdk_pixbuf_composite( image, bgImage, calc.x, calc.y, calc.width, calc.height,
							  0, 0, 1.0, 1.0, GDK_INTERP_NEAREST, 255 );
		gdk_draw_pixbuf( ( ( GtkWidget * ) drawable ) ->window, gc, bgImage,
						 0, 0, 0, 0, -1, -1, GDK_RGB_DITHER_NORMAL, 0, 0 );

		g_object_unref( splash );
		g_object_unref( image );
	}
	else if ( warnOnce )
	{
		warnOnce = false;
		modal_message( _( "Support pixmaps not found.\nPlease completely install Kino." ) );
	}

	g_object_unref( bgImage );
	g_object_unref( gc );
	free( background );
#endif
}


/** Make the video preview area black.
 
    \param widget the GtkDrawingArea to clear.
*/
void KinoCommon::clearPreview( GtkDrawingArea *drawable )
{
	GtkWidget * widget = GTK_WIDGET( drawable );
	if ( GDK_IS_DRAWABLE( widget->window ) )
	{
		gdk_draw_rectangle ( widget->window, widget->style->black_gc, TRUE,
		                     widget->allocation.x, widget->allocation.y,
		                     widget->allocation.width, widget->allocation.height );
	}
}


/** Set the video display to be fixed size or scalable.
 
    \param packerNameInner The name of the GTK AspectFrame widget to adjust.
    \param packerNameOuter The name of the GTK AspectFrame widget to adjust.
*/
void KinoCommon::packIt( const char *packerNameInner, const char* packerNameOuter )
{
	GdkColor color;
	gdk_color_parse ( "black", &color );

	GtkWidget *packer = lookup_widget( getWidget(), packerNameOuter );
	gtk_widget_modify_bg ( packer, GTK_STATE_NORMAL, &color );

	packer = lookup_widget( getWidget(), packerNameInner );
	gtk_widget_modify_bg ( packer, GTK_STATE_NORMAL, &color );

	GtkWidget *parent = gtk_widget_get_parent( packer );
	gtk_box_set_child_packing( GTK_BOX( parent ), packer,
	                           TRUE /* expand */,
	                           Preferences::getInstance().displayFixed ? FALSE : TRUE /* fill */,
	                           0 /* padding */, GTK_PACK_START );
}


/** Helper function for the pages */
void KinoCommon::setStatusBar( const char * msg, ... )
{
	va_list list;
	va_start( list, msg );
	static char prevMsg[ 1024 ] = "";

	if ( strcmp( msg, "" ) == 0 )
		gtk_statusbar_pop( statusbar, 1 );
	else if ( strncmp( prevMsg, msg, 1023 ) != 0 )
	{
		if ( vsnprintf( prevMsg, 1023, msg, list ) != 0 )
		{
			gtk_statusbar_pop( statusbar, 1 );
			gtk_statusbar_push( statusbar, 1, prevMsg );
		}
	}
}


bool KinoCommon::exitKino( )
{
	return newFile();
}

void KinoCommon::setTimeFormat( int format )
{
	timeFormat = format;
	if ( g_currentFrame != -1 )
	{
		common->moveToFrame();
		GetStoryboard() ->reset();
		GetStoryboard() ->redraw();
		if ( currentPage == PAGE_TIMELINE )
			selectScene( currentScene );
	}
}


string KinoCommon::makeTimecode( int i, int fps )
{
	char s[ 12 ];
	int hours, mins, secs;
	int cur = i;

	if ( i == 0 )
	{
		hours = 0;
		mins = 0;
		secs = 0;
	}
	else
	{
		/* NTSC drop-frame */
		if ( fps == 30 )
		{
			int max_frames = cur;
			for ( int j = 1800; j <= max_frames; j += 1800 )
			{
				if ( j != 18000 )
				{
					max_frames += 2;
					cur += 2;
				}
			}
		}
		hours = cur / ( fps * 3600 );
		cur -= hours * ( fps * 3600 );
		mins = cur / ( fps * 60 );
		cur -= mins * ( fps * 60 );
		secs = cur / fps;
		cur -= secs * fps;
	}

	snprintf( s, 12, "%2.2d:%2.2d:%2.2d%s%2.2d", hours, mins, secs, ( fps == 30 ) ? ";" : ":", cur );
	return string( s );
}

void KinoCommon::setCurrentScene( int frame )
{
	if ( frame > -1 )
	{
		int pos = 0;

		vector <int> scenes = getPageEditor() ->GetScene();
		for ( pos = 0; pos < ( int ) scenes.size() && frame >= scenes[ pos ]; pos++ )
			;
		if ( pos != currentScene && pos > -1 )
		{
			currentScene = pos;
			GetStoryboard() ->select( currentScene );
		}
	}
}


/*
 * Callbacks
 */

extern "C"
{

	gboolean
	on_open_movie_dialog_delete_event ( GtkWidget * widget,
	                                    GdkEvent * event,
	                                    gpointer user_data )
	{
		GtkWidget * dialog = gtk_widget_get_toplevel( widget );
		gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), "" );
		gtk_main_quit();
		return FALSE;
	}


	void
	on_open_movie_dialog_ok_button_clicked ( GtkButton * button,
	        gpointer user_data )
	{
		gtk_main_quit();
	}


	void
	on_open_movie_dialog_cancel_button_clicked
	( GtkButton * button,
	  gpointer user_data )
	{
		GtkWidget * dialog = gtk_widget_get_toplevel( GTK_WIDGET( button ) );
		gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), "" );
		gtk_main_quit();
	}


	gboolean
	on_save_movie_dialog_delete_event ( GtkWidget * widget,
	                                    GdkEvent * event,
	                                    gpointer user_data )
	{
		GtkWidget * dialog = gtk_widget_get_toplevel( widget );
		gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), "" );
		gtk_main_quit();
		return FALSE;
	}


	void
	on_save_movie_dialog_ok_button_clicked ( GtkButton * button,
	        gpointer user_data )
	{
		gtk_main_quit();
	}


	void
	on_save_movie_dialog_cancel_button_clicked
	( GtkButton * button,
	  gpointer user_data )
	{
		GtkWidget * dialog = gtk_widget_get_toplevel( GTK_WIDGET( button ) );
		gtk_file_selection_set_filename( GTK_FILE_SELECTION( dialog ), "" );
		gtk_main_quit();
	}

}
