/*
* storyboard.cc Storyboard view object
* Copyright (C) 2003 Dan Dennedy <dan@dennedy.org>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
using std::cout;
using std::endl;
#include <vector>
using std::vector;

#include <pthread.h>

#include "storyboard.h"
#include "frame.h"
#include "preferences.h"
#include "error.h"
#include "page_editor.h"

extern "C"
{
#include "support.h"
}
extern KinoCommon *common;

// Temporary(?) static variable to get round stack corruption(?) issue.
static unsigned char pixels[ FRAME_MAX_WIDTH*FRAME_MAX_HEIGHT*4 ];

enum SceneStatus {
    SCENE_INIT,
    SCENE_IDLE,
    SCENE_STARTING,
    SCENE_RUNNING,
    SCENE_RESTART
};
enum
{
    THUMBNAIL_COLUMN,
    START_COLUMN,
    N_COLUMNS
};

enum SceneStatus showScenesRunning = SCENE_INIT;
enum SceneStatus updateScenesRunning = SCENE_INIT;

static vector< GdkPixbuf * > *backup = NULL;
static pthread_mutex_t scene_mutex = PTHREAD_MUTEX_INITIALIZER;


void removeImages( vector< GdkPixbuf * >** list )
{
	if ( *list != NULL )
	{
		vector< GdkPixbuf * >::iterator iter;
		for ( iter = ( *list ) ->begin(); iter != ( *list ) ->end(); iter++ )
		{
			if ( *iter != NULL )
				g_object_unref( *iter );
		}
		( *list ) ->erase( ( *list ) ->begin(), ( *list ) ->end() );
		delete ( *list );
		*list = NULL;
	}
}

static void tree_model_row_inserted_cb( GtkTreeModel* model, GtkTreePath* path, GtkTreeIter* iter, gpointer data )
{
	Storyboard * storyboard = ( Storyboard* ) ( data );
	if ( ! storyboard->getSkip() )
	{
		gchar * row = gtk_tree_path_to_string( path );
		storyboard->moveScene( atoi( row ) );
		g_free( row );
	}
}

static gboolean tree_view_row_select( GtkWidget *widget, GdkEventButton *event, gpointer user_data )
{
	GtkTreeView * treeview = GTK_TREE_VIEW( widget );
	GtkTreeSelection *selection = gtk_tree_view_get_selection( treeview );

	GtkTreeIter iter;
	GtkTreeModel *model;

	if ( gtk_tree_selection_get_selected ( selection, &model, &iter ) )
	{
		gchar * row = gtk_tree_model_get_string_from_iter( model, &iter );
		common->selectScene( atoi( row ) );
		g_free( row );
	}

	return FALSE;
}

Storyboard::Storyboard( KinoCommon *common ) :
		common( common ), skipSelect( false ), skip( false )
{
	GtkCellRenderer * renderer;
	GtkTreeViewColumn* column;
	GtkTreeSelection* select;

	gdk_threads_enter();
	view = GTK_TREE_VIEW( lookup_widget( common->getWidget(), "treeview_storyboard" ) );
	g_signal_connect( G_OBJECT( view ), "button-release-event", G_CALLBACK( tree_view_row_select ), this );

	selection = common->getCurrentScene();
	model = gtk_list_store_new( N_COLUMNS, G_TYPE_OBJECT, G_TYPE_STRING );

	gtk_tree_view_set_model( view, GTK_TREE_MODEL( model ) );
	renderer = gtk_cell_renderer_pixbuf_new();
	column = gtk_tree_view_column_new_with_attributes(
	             _( "Storyboard" ), renderer, "pixbuf", THUMBNAIL_COLUMN, NULL );
	gtk_tree_view_column_set_fixed_width( column, 80 );
	gtk_tree_view_column_set_min_width( column, 80 );
	gtk_tree_view_column_set_max_width( column, 80 );
	gtk_tree_view_append_column( view, column );
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(
	             _( "File / Begin" ), renderer, "text", START_COLUMN, NULL );
	gtk_tree_view_append_column( view, column );

	select = gtk_tree_view_get_selection( view );
	gtk_tree_selection_set_mode( select, GTK_SELECTION_SINGLE );
	g_signal_connect( G_OBJECT( model ), "row-inserted",
	                  G_CALLBACK( tree_model_row_inserted_cb ), static_cast<gpointer>( this ) );

	gdk_threads_leave();
	g_object_unref( G_OBJECT( model ) ); // view has the reference
}

Storyboard::~Storyboard()
{}

void *showScenesThread( void *arg )
{
	Frame & frame = *( GetFramePool( ) ->GetFrame( ) );
	GdkPixbuf *image = NULL;
	GdkPixbuf *thumbnail = NULL;
	static vector< GdkPixbuf * > *icons = NULL;
	vector <int> scene;
	int frameNum = 0;
	Storyboard *storyboard = static_cast<Storyboard *>( arg );
	GtkTreeIter iter;

	frame.decoder->quality = DV_QUALITY_DC;
	if ( Preferences::getInstance().displayQuality < 4 )
		frame.decoder->quality |= DV_QUALITY_COLOR;


	while ( showScenesRunning != SCENE_IDLE )
	{
		// Apparently(?) the thumbs generated cannot be removed while they're used in the
		// icon list - this doesn't appear to be the case for the timeline page, but is
		// definitely causing problems here - could be the modified images causing an issue
		// perhaps? Anyway, to avoid a blank list during the execution of this thread, take
		// a backup of the previous list and create a new one (see comments below).

		icons = new vector< GdkPixbuf *>;

		// Change from SCENE_STARTING to SCENE_RUNNING
		showScenesRunning = SCENE_RUNNING;

		// Get the scene list (contains the frame number first frame of each scene)
		scene = common->getPageEditor() ->GetScene();

		// Create the icons vector but terminate if the running state changes
		for ( unsigned int i = 0; showScenesRunning == SCENE_RUNNING && i < scene.size(); ++i )
		{

			frameNum = i == 0 ? 0 : scene[ i - 1 ];

			fail_neg( common->getPlayList() ->GetFrame( frameNum, frame ) );

			frame.ExtractHeader();
			frame.ExtractRGB( pixels );

			// Get the image
			image = gdk_pixbuf_new_from_data( pixels, GDK_COLORSPACE_RGB, FALSE, 8,
			                                  frame.GetWidth(), frame.GetHeight(), frame.GetWidth() * 3,
			                                  NULL, NULL );

			// create the thumbnail
			// TODO: apply PAL aspect ratio factor
			thumbnail = gdk_pixbuf_scale_simple( image, 80, 60, GDK_INTERP_NEAREST );
			icons->push_back( thumbnail );
			g_object_unref( image );
		}

		// Render the icons if we're still running, otherwise remove the [partial] list generated
		// NB: The freeze followed by a thread_leave causes the icon list to be blanked on the screen and
		// removing the current list before clearing the iconlist causes possible core dumps if a redraw
		// is required before the new list is in place - hence the backup keeps the last used list until
		// we're sure that we can safely remove them

		if ( showScenesRunning == SCENE_RUNNING )
		{
			pthread_mutex_lock( &scene_mutex );
			storyboard->setSkip();
			gdk_threads_enter();
			gtk_list_store_clear( storyboard->getModel() );
			for ( unsigned int i = 0; showScenesRunning == SCENE_RUNNING && i < scene.size(); ++i )
			{
				char start[ 1024 ] = "";

				frameNum = i == 0 ? 0 : scene[ i - 1 ];
				if ( common->getTimeFormat() == TIME_FORMAT_SMPTE )
				{
					fail_neg( common->getPlayList() ->GetFrame( frameNum, frame ) );
					frame.ExtractHeader();
					snprintf( start, 1024, "%s\n%s",
					          basename( common->getPlayList() ->GetFileNameOfFrame( frameNum ) ),
					          common->makeTimecode( frameNum, static_cast<int>( frame.GetFrameRate() + 0.5 ) ).c_str() );
				}
				else if ( common->getTimeFormat() == TIME_FORMAT_FRAMES )
				{
					snprintf( start, 1024, "%s\n%d",
					          basename( common->getPlayList() ->GetFileNameOfFrame( frameNum ) ), frameNum );
				}

				gtk_list_store_append( storyboard->getModel(), &iter );
				gtk_list_store_set( storyboard->getModel(), &iter,
				                    THUMBNAIL_COLUMN, ( *icons ) [ i ],
				                    START_COLUMN, start,
				                    -1 );
			}
			storyboard->select( common->getCurrentScene() );
			gdk_threads_leave();
			storyboard->clearSkip();
			pthread_mutex_unlock( &scene_mutex );
			// Clean up temporary backup now
			removeImages( &backup );
			// Now backup
			backup = icons;
		}
		else
		{
			// Remove any unused thumbs in icons
			removeImages( &icons );
		}
		// If we're still running now, it's OK to idle again.
		if ( showScenesRunning == SCENE_RUNNING )
			showScenesRunning = SCENE_IDLE;

	}
	GetFramePool( ) ->DoneWithFrame( &frame );

	return NULL;
}

void Storyboard::redraw( )
{
	static pthread_t scenes_thread;
	static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

	pthread_mutex_lock( &mutex );
	selection = -1;

	// If running then, set as RESTART otherwise start the thread,
	// otherwise we're already waiting on a restart so do nothing
	if ( showScenesRunning == SCENE_RUNNING )
	{
		showScenesRunning = SCENE_RESTART;
	}
	else if ( showScenesRunning == SCENE_INIT )
	{
		showScenesRunning = SCENE_STARTING;
		pthread_create( &scenes_thread, NULL, showScenesThread, this );
	}
	else if ( showScenesRunning == SCENE_IDLE )
	{
		pthread_join( scenes_thread, NULL );
		showScenesRunning = SCENE_STARTING;
		pthread_create( &scenes_thread, NULL, showScenesThread, this );
	}

	pthread_mutex_unlock( &mutex );
}

void Storyboard::select( int scene )
{
	if ( scene < gtk_tree_model_iter_n_children( GTK_TREE_MODEL( model ), NULL ) )
	{
		GtkTreeIter iter;
		GtkTreeSelection* selection = gtk_tree_view_get_selection( view );
		if ( gtk_tree_model_iter_nth_child( GTK_TREE_MODEL( model ), &iter, NULL, scene ) )
		{
			GtkTreePath * path = gtk_tree_model_get_path( GTK_TREE_MODEL( model ), &iter );
			skipSelect = true;
			gtk_tree_selection_select_iter( selection, &iter );
			gtk_tree_view_scroll_to_cell( view, path, NULL, TRUE, 0.5, 0.0 );
			gtk_tree_path_free( path );
			skipSelect = false;
		}
		this->selection = scene;
	}
}

void Storyboard::setSensitive( bool sensitive )
{
	gtk_widget_set_sensitive( GTK_WIDGET( view ), sensitive ? TRUE : FALSE );
}

void Storyboard::moveScene( unsigned int destScene )
{
	cout << "moving scene " << selection << " to scene " << destScene << endl;
	vector <int> scenes = common->getPageEditor() ->GetScene();
	int start = ( selection == 0 ) ? 0 : scenes[ selection - 1 ];
	int end = common->getPlayList() ->FindEndOfScene( start );
	int destFrame = 0;

	common->getPageEditor() ->CopyFrames( start, end );
	common->getPlayList() ->Delete( start, end );

	if ( destScene >= scenes.size() )
	{
		destFrame = common->getPlayList() ->GetNumFrames();
	}
	else if ( ( int ) destScene > selection )
	{
		common->getPageEditor() ->ResetBar();
		scenes = common->getPageEditor() ->GetScene();
		destFrame = scenes[ destScene - 2 ];
	}
	else if ( destScene > 0 )
	{
		common->getPageEditor() ->ResetBar();
		scenes = common->getPageEditor() ->GetScene();
		destFrame = scenes[ destScene - 1 ];
	}

	common->getPageEditor() ->PasteFrames( destFrame );
	common->moveToFrame( destFrame );

	redraw( );
}

Storyboard *GetStoryboard( )
{
	static Storyboard * singleton = new Storyboard( common );
	return singleton;
}
